/*Copyright (c) Dr. Stephen Ware and the University of New Orleans. All rights reserved.
Contributors: Alireza Shirvani, Edward T. Garcia
Licensed under the Non-Profit Open Software License version 3.0. See LICENSE file in the project root for full license information.*/
using Medium;
using PlotGraph;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ExperienceManager
{
    public class ExperienceManager
    {
        internal readonly Graph plotGraph;
        internal readonly Database database;
        private static ExperienceManager instance;

        string player;
        ProcessCore core;

        State state;
        public enum State
        {
            none, initialized, started, ready
        }

        public ExperienceManager(Graph plotGraph, Database database)
        {
            instance = this;
            state = State.none;
            this.plotGraph = plotGraph;
            this.database = database;
            core = new ProcessCore();
            currentNPCProcessors = new List<SequenceProcessor>();
        }

        public void Run(string player)
        {
            this.player = player;
#if DEBUG
            RunProcess();
#else
            Initialize(new InitialState().Get());
#endif
            core.Process();
        }

        private void Initialize(List<string> list)
        {
            database.Initialize(list);
            var sequence = new SequenceProcessor(core, list, null, (processor) =>
            {
                state = State.started;
                Console.WriteLine("start Game");
                Logger.WritePlot("start Game");
                RunProcess();
            });

            state = State.initialized;
        }

        SequenceProcessor currentPlayerProcessor;
        List<SequenceProcessor> currentNPCProcessors;

        //Determines if the action is a player action
        //Here, a player action starts with the player as the first parameter
        bool IsPlayerAction(string action)
        {
            return action.ToLower().Substring(action.IndexOf('(') + 1).StartsWith(player.ToLower());
        }

        List<string> PlayerActions(List<string> actions)
        {
            return actions.Where(z => IsPlayerAction( z)).ToList();
        }

        public List<string> NPCActions(List<string> actions)
        {
            return actions?.Where(z => !IsPlayerAction(z)).ToList();
        }
        
        void RunProcess()
        {
            var iprocessor = new InputProcessor(core);

            iprocessor.OnInput += (action) =>
            {
                //Console.WriteLine("@Received input: " + action);
                if ((currentPlayerProcessor == null || currentPlayerProcessor.state != SequenceProcessor.ProcessorState.Running)
                && waitingNPC.Where(z=>z.ToString().Contains(player)).Count() == 0)
                {
                    var actions = plotGraph.Actions();
                    var playerActions = PlayerActions(actions);
                    //Console.WriteLine("@"+playerActions.Aggregate((a, b) => a + "," + b));
                    var sequence = database.Get(action.ToString(), playerActions.ToList());
                    if (sequence != null)
                    {
                        if (currentPlayerProcessor != null)
                        {
                            currentPlayerProcessor.Interrupt();
                        }
                        currentPlayerProcessor = new SequenceProcessor(core, sequence, OnCommitted, OnCompleted, true);
                    }
                    else
                    {
                        //Console.WriteLine("@Received input: " + action + " but no such sequence present.");
                    }
                }
                else
                {
                    if(currentPlayerProcessor!=null && currentPlayerProcessor.state == SequenceProcessor.ProcessorState.Running)
                    {
                        //Console.WriteLine("@Received input: " + action + " but busy doing: " + currentPlayerProcessor.ToString());
                    }else if(waitingNPC.Where(z => z.ToString().Contains(player)).Count()> 0)
                    {
                        //Console.WriteLine("@Received input: " + action + " but waiting for: " + waitingNPC.Select(z=>z.ToString()).Aggregate((a,b)=>a+" , "+b));
                    }
                }
            };
        }

        List<System.Action> waitingToRun = new List<System.Action>();
        internal static void AskPermission(System.Action run)
        {
            if (instance.waitingNPC.Count() > 0 || instance.waitingPlayer)
                instance.waitingToRun.Add(run);
            else
                run();

        }

        IEnumerable<SequenceProcessor> waitingNPC
        {
            get
            {
                return currentNPCProcessors.Where(z => z.state == SequenceProcessor.ProcessorState.Committed);
            }
        }

        bool waitingPlayer
        {
            get
            {
                return currentPlayerProcessor.state == SequenceProcessor.ProcessorState.Committed;
            }
        }

        static SequenceProcessor committed { get; set; }
        
        static bool GetPermission(SequenceProcessor sp)
        {
            return committed == null || committed.Equals(sp);
        }

        internal static bool SetCommitted(SequenceProcessor sequenceProcessor)
        {
            if (committed == null)
            {
                committed = sequenceProcessor;
                return true;
            }
            return false;
        }

        public static bool GetPermission(bool player, SequenceProcessor sp)
        {
            return committed == null || 
                (!player && GetPermission(sp)) || 
                (player && (!instance.ToString().Contains(instance.player) || GetPermission(sp)));
        }

        void OnCommitted(SequenceProcessor sp)
        {
            if (sp.ToString().Equals("input go(Home.Door)"))
                state = State.ready;

            UpdateState(sp.ToString(), player);

            var actions = plotGraph.Actions();
            database.Update(PlayerActions(actions)).ForEach(z => {
                Console.WriteLine("start " + z);
                Logger.WritePlot("start " + z);
            });
            var npcActions = NPCActions(actions);
            foreach (var item in npcActions)
            {
                var msg="@next NPC action: " + item + "' with priority " + plotGraph.GetPriority(item.ToString());
                Console.WriteLine(msg);
                Logger.WritePlot(msg);
            }
            currentNPCProcessors.Where(z=>z.state== SequenceProcessor.ProcessorState.Running).ToList().ForEach(z =>
            {
                if (!npcActions.Contains(z.ToString()))
                {
                    var msg="@interrupted action: " + z.ToString();
                    Console.WriteLine(msg);
                    Logger.WritePlot(msg);
                    z.Interrupt();
                }
            });
            currentNPCProcessors.RemoveAll(z => z.state == SequenceProcessor.ProcessorState.Completed);
            
        }

        void OnCompleted(SequenceProcessor sp)
        {
            if (sp.Equals(committed))
            {
                committed = null;
            }

            currentNPCProcessors.Remove(sp);
            var actions = NPCActions(plotGraph.Actions());
            foreach (var item in waitingNPC)
            {
                var msg = "@waiting for NPC action to complete: '" + item ;
                Console.WriteLine(msg);
                Logger.WritePlot(msg);
            }
            if (waitingNPC.Count() == 0 && state == State.ready && committed==null)
            {
                PerformNPCActions(actions);
            }
        }

        private void PerformNPCActions(List<string> actions)
        {
            foreach (var NPCAction in actions.OrderByDescending(z=>plotGraph.GetPriority(z)))
            {
                if (currentNPCProcessors.Any(z => z.ToString().Equals(NPCAction)))
                    continue;
                if (waitingPlayer && NPCAction.Contains(player))
                {
                    var msg = "@waiting for player action to complete: " + currentPlayerProcessor.ToString() + " for: " + NPCAction;
                    Console.WriteLine(msg);
                    Logger.WritePlot(msg);
                    continue;
                }
                Console.WriteLine("@###" + NPCAction + plotGraph.GetPriority(NPCAction));
                var sequence = new Sequence(database.Get(NPCAction));
                if (sequence != null)
                {
                    currentNPCProcessors.Add(new SequenceProcessor(core, sequence, OnCommitted, OnCompleted));
                }
            }
        }

        bool UpdateState(string action, string player)
        {
            int before = plotGraph.Current();
            var current = plotGraph.Navigate(action);
            if(current.HasValue)
            {
                var msg = "@Plotgraph state: from " + before + " to " + current.Value + " after " + action;
                Console.WriteLine(msg);
                Logger.WritePlot(msg);
                return true;
            }
            return false;
        }
    }
}


