/*Copyright (c) Dr. Stephen Ware and the University of New Orleans. All rights reserved.
Contributors: Alireza Shirvani, Edward T. Garcia
Licensed under the Non-Profit Open Software License version 3.0. See LICENSE file in the project root for full license information.*/
using Medium;
using PlotGraph;
using System;
using System.Collections.Generic;

namespace ExperienceManager
{
    public class SequenceProcessor : Processor
    {
        int current;
        Sequence sequence;
        internal System.Action<SequenceProcessor> OnCommitted { get;  set; }
        internal System.Action<SequenceProcessor> OnCompleted { get; set;}
        UnityAction current_action;
        Queue<UnityAction> actions;
        List<UnityAction> completedActions;
        public ProcessorState state { get; private set; }
        public enum ProcessorState
        {
            None, Running, Committed, Completed
        }

        SequenceProcessor(ProcessCore core, System.Action<SequenceProcessor> committed, System.Action<SequenceProcessor> completed) : base(core)
        {
            state = ProcessorState.None;
            this.OnCompleted = completed;
            this.OnCommitted = committed;
            completedActions = new List<UnityAction>();
        }

        public SequenceProcessor(ProcessCore core, List<string> list, System.Action<SequenceProcessor> committed, System.Action<SequenceProcessor> completed) : this(core,committed, completed)
        {
            if (list != null)
            {
                actions = new Queue<UnityAction>();
                completedActions = new List<UnityAction>();
                foreach (var item in list)
                {
                    var uaction = new UnityAction(item.Replace("*", "").Trim());
                    uaction.dontwait = (!item.Contains("*") ? null : (bool?)false);
                    actions.Enqueue(uaction);
                }
                Run();
            }
            else
            {
                Completed();
            }
        }

        protected override void Completed()
        {
            state = ProcessorState.Completed;
            if (OnCompleted != null)
                OnCompleted(this);
            base.Completed();
        }

        bool player;
        public SequenceProcessor(ProcessCore core, Sequence sequence, System.Action<SequenceProcessor> committed, System.Action<SequenceProcessor> completed, bool player=false) : this(core,committed, completed )
        {
            this.player = player;
            if (sequence == null)
            {
                Completed();
            }
            else
            {
                current = 0;
                actions = new Queue<UnityAction>();
                this.sequence = sequence;
                Run();
            }
        }

        protected override void Process(Message message)
        {
            if (message is StatusMessage && message.action.Replace("'", string.Empty).Equals(current_action.action.Replace("'", string.Empty)))
            {
                switch (((StatusMessage)message).status)
                {
                    case (Status.succeeded):
                        current++;
                        if (current == sequence?.primary)
                        {
                            if(ExperienceManager.SetCommitted(this))
                            Console.WriteLine("got committed " + this.ToString());
                        }
                        if (current != sequence?.primary || ExperienceManager.GetPermission(player, this))
                        {
                            Run();
                        }
                        else
                        {
                            Completed();
                        }
                        break;
                    case (Status.started):
                        if (current == sequence?.primary)
                        {
                            Committed();
                        }
                        if (current_action.dontwait.HasValue && current_action.dontwait.Value)
                        {
                            Run();
                        }
                        break;
                    case (Status.failed):

                        if (state == ProcessorState.Committed)
                            throw new Exception("Oops! Action failed after being committed: '"+ToString()+"'");
                        
                        Failed();

                        break;
                }
            }
        }

        void Run()
        {
            if (actions.Count == 0)
            {
                var list = this.sequence?.Next();
                if (list != null)
                    list.ForEach(z => actions.Enqueue(z));
            }
            if (actions.Count > 0)
            {
                current_action = actions.Dequeue();
                completedActions.Add(current_action);
                if (state == ProcessorState.None)
                    state = ProcessorState.Running;
                var msg = string.Format("start {0}", current_action.action);
                Console.WriteLine(msg);
                Logger.WritePlot(msg);
                if (!current_action.dontwait.HasValue)
                    Run();
            }
            else
            {
                Committed();
                Completed();
            }
        }

        private void Committed()
        {
            if (state != ProcessorState.Committed)
            {
                state = ProcessorState.Committed;
                if (OnCommitted != null)
                    OnCommitted(this);
            }
        }

        protected override void Failed()
        {
            if (state == ProcessorState.Running)
            {
                state = ProcessorState.Completed;
                if (completedActions.Count > 0)
                    for (int i = completedActions.Count - 1; i >= 0; i--)
                    {
                        Console.WriteLine("@stopped " + completedActions[i].action);
                        Console.WriteLine("stop {0}", completedActions[i].action);
                        base.Failed();
                    }
            }
        }

        public override string ToString()
        {
            return sequence.ToString();
        }
    }
}


