/*Copyright (c) Dr. Stephen Ware and the University of New Orleans. All rights reserved.
Contributors: Alireza Shirvani, Edward T. Garcia
Licensed under the Non-Profit Open Software License version 3.0. See LICENSE file in the project root for full license information.*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;

namespace Medium
{
    public class Database
    {
        Lists lists;
        ActionSequences actions;
        Icons icons;
        List<Item> items;
        Fields fields;
        Bindings bindings;
        Fluents fluents;
        Sequence actionNotInPlotgraph;
        internal static Strings strings;

        List<string> playerActions;

        public Sequence Get(string input, List<string> playerActions = null)
        {
            this.playerActions = playerActions;
            if (playerActions != null && bindings.Contains(input.ToLower()))
            {
                var match = bindings[input];
                var action = playerActions.FirstOrDefault(z => match(z.ToLower()));
                if (action != null)
                    return actions.Get(action);
            }
            return actions.Get(input);
        }

        public Database()
        {
            strings = new Strings();
            fluents = new Fluents();
            fluents.Set("PlayerLocation", "Market".GetString());

            icons = new Icons();
            fields = new Fields();
            items = Items();
            lists = new Lists();
            lists.Add("Merchant".GetString(), GetItem("Medicine".GetString()));
            lists.Add("Merchant".GetString(), GetItem("MerchantSword".GetString()));
            lists.Add("Grandmother".GetString(), GetItem("PlayerCoin".GetString()));
            lists.Add("Guard".GetString(), GetItem("GuardSword".GetString()));
            lists.Add("Bandit".GetString(), GetItem("BanditSword".GetString()));

            lists.Add("Camp.Chest".GetString(), GetItem("BanditCoin".GetString()));
            bindings = GetBindings();
            actions = GetAll();

            actionNotInPlotgraph = new Sequence("default", "empty")
                .Add("SetNarration", false, "'You cannot do this action at this time.'".GetString())
                .Add("ShowNarration");

        }

        public void Initialize(List<string> list)
        {
            var positionActions = list.Where(z => z.StartsWith("SetPosition"));
            foreach (var action in positionActions)
            {
                var captures = Regex.Matches(action, @"[^\(,\)]+");
                if (captures.Count > 0)
                {
                    var args = GetArgs(captures).ToList();
                    if (args[0].Equals("Player"))
                        fluents.Set("PlayerLocation", args[1]);
                    else if (GetItem(args[0]) != null)
                    {
                        lists.Add(args[1], GetItem(args[0]));
                    }
                }
            }

            var iconActions = list.Where(z => z.StartsWith("EnableIcon"));
            foreach (var action in iconActions)
            {
                var captures = Regex.Matches(action, @"[^\(,\)]+");
                if (captures.Count > 0)
                {
                    var args = GetArgs(captures).ToList();
                    icons.Enable(args[2], args[0], defaultIcon: args.Count > 4 && args[4].ToLower().Equals("true"));
                }
            }
        }

        private IEnumerable<string> GetArgs(MatchCollection captures)
        {
            for (int i = 1; i < captures.Count; i++)
            {
                yield return captures[i].Value.Trim();
            }
        }

        /// <summary>
        /// Commands that start with input are either only unity actions or domain actions that are not possible in the current node of the plotgraph
        /// </summary>
        /// <returns></returns>
        private ActionSequences GetAll()
        {
            return new ActionSequences(this)
            {
                new Sequence("input Open","Cottage.Door").If((d) =>
                    d.lists.Contains("Player", "Medicine".GetString())
                )
                    .Add((a,d)=> Win(d)),

                new Sequence("input Key Restart")
                .Add((a, d)=>Reset(d)),

                new Sequence("input Open","Cottage.Door").If((d) =>
                    !d.lists.Contains("Player", "Medicine".GetString())
                )
                    .Add("SetNarration", false, "'I have to get my grandmother`s medicine first!'".GetString())
                    .Add("ShowNarration"),

                new Sequence("input Selected Start")
#if DEBUG
#else
                     .Add("SetCameraMode", false, "follow")
                     .Add("SetCameraFocus", false, "Player")
                    .Add("FadeIn").DontWait()
                    .Add("EnableInput").DontWait()
                    .Add("SetNarration",false,"'LEFT click on a highlighted character or object to perform the default action. RIGHT click to choose an action.'".GetString())
#endif
                    .Add("ShowNarration"),

                new Sequence("input Cancel").If((d)=>{return d.fluents.IsTrue("GameOver"); })
                    .Add("Quit"),

                new Sequence("input Key Cancel").If((d)=>{return d.fluents.IsTrue("GameOver"); })
                    .Add("Quit"),

                new Sequence("input Cancel").If((d)=>{return d.fluents.IsFalse("GameOver"); })
                    .Add("ClearList")
                    .Add((a,d)=>{d.fluents.SetFalse("InventoryOpen"); })
                    .Add("HideNarration")
                    .Add("HideList"),

                new Sequence("input Key Cancel").If((d)=>{return d.fluents.IsFalse("GameOver") && d.fluents.IsTrue("InventoryOpen");  })
                    .Add("ClearList")
                    .Add((a,d)=>{d.fluents.SetFalse("InventoryOpen"); })
                    .Add("HideList"),

                new Sequence("input Key Cancel").If((d)=>{return d.fluents.IsFalse("GameOver") && d.fluents.IsFalse("InventoryOpen"); })
                    .Add("HideNarration"),

                new Sequence("input Key Inventory").If((d)=>{return d.fluents.IsFalse("InventoryOpen"); })
                    .Add("ClearList")
                    .Add((a,d)=>{return d.lists.GetAddToListActions("Player"); })
                    .Add("ShowList",true, "Player", "Inventory".GetString())
                    .Add((a,d)=>{d.fluents.SetTrue("InventoryOpen"); }),

                new Sequence("input Key Inventory").If((d)=>{return d.fluents.IsTrue("InventoryOpen"); })
                    .Add("HideList")
                    .Add((a,d)=>{d.fluents.SetFalse("InventoryOpen"); }),

                new Sequence("Walk", "character", "from", "fromDoor", "toDoor", "to").If((d,args)=>{return args[0].Equals("Player"); })
#if DEBUG
#else
                    .Add("WalkTo", false, "Player", "fromDoor")
                    .Add("DisableInput", true)
                    .Add((a,d)=>{return string.Format("Exit({0}, {1}, true)",a[0],a[1]); },true,"Player", "fromDoor")
                    .Add((a,d)=>{
                        return true?//a[1].Equals("Market.Gate")||a[1].Equals("Crossroads.Gate")?
                        string.Format("Enter({0}, {1}, true)",a[0],a[1]):""; }
                    ,false,"Player", "toDoor")
                    //.Add("SetPosition",false, "Player", "toDoor")
                    .Add((a,d)=>{d.fluents.Set("PlayerLocation", a[0]);}, "to")
                    .Add("FadeIn").DontWait()
#endif
                .Add("EnableInput"),

                new Sequence("Walk", "character", "from", "fromDoor", "toDoor", "to").If((d,args)=>{return !args[0].Equals("Player"); })
                    .Add("WalkTo", false, "character", "fromDoor")
                    .Add((a,d)=>{return string.Format("Exit({0}, {1})",a[0],a[1]); },true,"character", "fromDoor")
                    .Add((a,d)=>{return string.Format("Enter({0}, {1})",a[0],a[1]); },false,"character", "toDoor"),

                new Sequence("Rob", "thief", "item", "owner", "place")
                    .Add((a, d) =>
                    {
                        var list=new List<String>();
                        if(!a[0].Equals("Player"))
                            list.Add(string.Format("WalkTo({0}, {1})", a[0], a[1]));
                        if(a[0].Equals("Player")||a[1].Equals("Player"))
                            list.Add("HideList()");
                        return list;
                    }, false, "thief", "owner")
                    .Add("StopMovement", false, "owner")
                    .Add((a,d)=>{
                        var list=new List<string>();
                        if (a[0].Equals("Player")||a[1].Equals("Player"))
                           list.Add("DisableInput()");
                        list.Add(string.Format("Draw({1}, {0})",d.lists.Get("name", a[0], "sword", bool.TrueString),a[0]));
                        return list;
                    },true,"thief", "owner")
                    .Add("SetExpression", false, "owner", "'Scared'")
                    .Add("Take", false, "thief", "item", "owner")
                    .Add((a,d)=>{ return d.icons.DisableAll(a[0]); },false,"item")
                    .Add((a,d)=>{ d.lists.Add(a[0], GetItem(a[1])); }, "thief", "item")
                    .Add((a,d)=>{ d.lists.Remove(a[0],GetItem(a[1])); }, "owner", "item")
                    .Add((a,d)=>{
                        var item=d.GetItem(a[0]);
                        if(item==null)
                            throw new PlotGraph.PlotGraphException(string.Format("item {0} does not exist", a[0]));
                        return d.GetItem(a[0])["Own"]; },false,"item")
                    .Add("Pocket", false, "thief", "item")
                    .Add((a,d)=>{
                        if (a[0].Equals("Player"))
                            return new List<String>(){
                                string.Format("SetNarration('The {0} stole your {1}.')".GetString(),a[2], d.GetItem(a[1])["description"]?.ToLower()),
                                "ShowNarration()"
                            };
                        return null;
                    },false,"owner", "item", "thief")
                    .Add((a,d)=>{return string.Format("Sheathe({1}, {0})",d.lists.Get("name", a[0], "sword", bool.TrueString),a[0]); },false, "thief")
                    .Add((a,d)=>{
                        if (a[0].Equals("Player")||a[1].Equals("Player"))
                           return "EnableInput()";
                        return null;
                    },false,"owner", "thief")
                    .Add("SetExpression", false, "owner", "'Neutral'"),

                new Sequence("Rob", "thief", "item", "owner").If((d,args)=>{
                        return d.CharacterIsArmed(args[0]);
                    })
                    .Add("WalkTo", false,"thief", "owner")
                    .Add((a,d)=>{return string.Format("Draw({1}, {0})",d.lists.Get("name", a[0], "sword", bool.TrueString),a[0]); },true, "owner"),

                new Sequence("input Rob", "owner").If((d,args)=>{
                        return d.CharacterIsArmed(args[0]);
                    })
                    .Add("WalkTo", false,"Player", "owner")
                    .Add((a,d)=>{return string.Format("Draw({1}, {0})",d.lists.Get("name", a[0], "sword", bool.TrueString),a[0]); },true, "owner"),

                new Sequence("input Steal", "item")
                    .Add("SetNarration", false,"'You cannot steal this item at this time.'".GetString())
                    .Add("ShowNarration"),

                new Sequence("input Rob", "robbed").If((d,args)=>{
                        return d.CharacterIsArmed("Player")&&
                        !d.CharacterIsArmed(args[0]);
                    })
                    .Add("WalkTo", false, "Player", "robbed")
                    .Add("ClearList", true)
                    .Add((a,d)=>{d.fluents.Set("robbed", a[0]); }, "robbed")
                    .Add((a,d)=>{return d.lists.GetAddToListActions(a[0]); },false,"robbed")
                    .Add((a,d)=>{return d.lists.Get(a[0])?.Select(z=> icons.Enable(z["name"], "Steal", "Steal", "Steal "+z["description"],true)).ToList(); },false,"robbed")
                    .Add("ShowList",true, "robbed", "'Items to steal'".GetString()),

                new Sequence("input Trade", "other")
                    .Add("WalkTo", false,"Player", "other")
                    .Add("SetLeft",false, "Player")
                    .Add("SetRight",false, "other")
                    .Add("SetDialog",false, "'Take a look at my fine wares! [ShowMerchandise|Ok] [Nevermind|Never mind]'".GetString())
                    .Add("ShowDialog", true),

                new Sequence("input Report", "other")
                    .If((d) =>
                    {
                        return playerActions==null||!playerActions.Any(z=>z.Contains("report"));
                    })
                    .Add("WalkTo", false,"Player", "other")
                    .Add("SetLeft",false, "Player")
                    .Add("SetRight",false, "other")
                    .Add("SetDialog('What is it, young man? [Nevermind|Never mind]')".GetString())
                    .Add("ShowDialog", true),

                new Sequence("input Report", "other")
                    .If((d) =>
                    {
                        return playerActions!=null&& playerActions.Any(z=>z.Contains("report"));
                    })
                    .Add("WalkTo", false,"Player", "other")
                    .Add("SetLeft",false, "Player")
                    .Add("SetRight",false, "other")
                    .Add("SetDialog('What is it, young man? [Reported|Tell the guard about the bandit] [Nevermind|Never mind]')".GetString())
                    .Add("ShowDialog", true),

                new Sequence("Report", "character", "other", "target", "place")
                    .Add((a,d)=>{
                    if(a[0].Equals("Player"))
                            return "HideDialog()";
                    else
                            return string.Format("WalkTo({0}, {1})", a[0], a[1]);
                    }, false, "character", "other")
                    .Add("SetExpression", false, "character", "'Neutral'"),

                new Sequence("input Selected Nevermind")
                    .Add("HideDialog"),

                new Sequence("input Selected ShowMerchandise")
                    .Add("HideDialog")
                    .Add("ClearList")
                    .Add((a,d)=>{return d.lists.GetAddToListActions(a[0]); },false,"Merchant")
                    .Add((a,d)=>{return d.lists.Get(a[0])?.Select(z=> d.icons.Enable(z["name"], "Buy", "Buy", "Buy "+z["name"],true)).ToList(); },false,"Merchant".GetString())
                    .Add((a,d)=>{return d.lists.Get(a[0])?.Select(z=> d.icons.Enable(z["name"], "Steal", "Steal", "Steal "+z["name"],false)).ToList(); },false,"Merchant".GetString())
                    .Add("ShowList",true, "Merchant".GetString(), "'Merchandise'".GetString()),

                new Sequence("Buy", "buyer", "item", "Merchant", "price", "place").If((d,args)=>{
                        return d.lists.Contains(args[0], args[3]);
                    })
                    .Add("WalkTo", false,"buyer", "Merchant".GetString())
                    .Add("SetPosition", true, "item","nowhere")
                    .Add((a,d)=>(a[0].Equals("Player")?"HideList()":null),false, "buyer")
                    .Add("Give", false, "Merchant".GetString(), "item", "buyer")
                    .Add((a,d)=>d.icons.DisableAll(a[0]),false, "item")
                    .Add((a,d)=>{ d.lists.Remove(a[0],d.GetItem(a[1])); }, "Merchant".GetString(), "item")
                    .Add((a,d)=>{ d.lists.Add(a[0], d.GetItem(a[1])); },"buyer", "item")
                    .Add((a,d)=>{d.lists.Remove(a[0],d.GetItem(a[1]));}, "buyer", "price")
                    .Add((a,d)=>{ d.lists.Add(a[0], d.GetItem(a[1])); },"Merchant".GetString(), "price")
                    .Add((a,d)=>{
                        var item=d.GetItem(a[0]);
                        if(item==null)
                            throw new PlotGraph.PlotGraphException(string.Format("item {0} does not exist", a[0]));
                        return d.GetItem(a[0])["Own"]; },false,"item")
                    .Add((a,d)=>{ return d.icons.DisableAll(a[0]); },false,"item")
                    .Add("Pocket", false, "buyer", "item"),

                new Sequence("input Buy","item").If((d,args)=>{
                        return !d.lists.Contains("Player", GetItem("PlayerCoin".GetString()))&&!d.lists.Contains("Player", GetItem("BanditCoin".GetString()));
                    })
                    .Add("SetNarration",false,"'You cannot pay for this item.'".GetString())
                    .Add("ShowNarration",true),

                new Sequence("input Buy","item").If((d,args)=>{
                        return d.lists.Contains("Player", GetItem("PlayerCoin".GetString()))||d.lists.Contains("Player", GetItem("BanditCoin".GetString()));
                    })
                    .Add("SetNarration",false,"'You cannot buy this item at this time.'".GetString())
                    .Add("ShowNarration",true),

                 new Sequence("input Buy","item").If((d,args)=>{
                        return d.lists.Contains("Player", args[3]);
                    })
                    .Add("SetNarration",false,"'You cannot buy this item at this time.'".GetString())
                    .Add("ShowNarration",true),

                new Sequence("Attack","attacker","victim", "place").If((d,args)=>{
                        return !d.CharacterIsArmed(args[0]) && d.CharacterIsArmed(args[1]);
                    })
                    .Add((a,d)=>{return string.Format("Draw({1}, {0})",d.lists.Get("name", a[0], "sword", bool.TrueString),a[0]); },true, "victim"),

                new Sequence("input Attack","victim")
                    .Add("SetNarration",false,"'You cannot attack this person at this time.'".GetString())
                    .Add("ShowNarration",true),

                new Sequence("Attack","attacker","victim", "place").If((d,args)=>{
                        return d.CharacterIsArmed(args[0]);
                    })
                    .Add("WalkTo",false, "attacker", "victim")
                    .Add("StopMovement", false, "victim")
                    .Add((a,d)=>{
                        var list=new List<string>();
                        if(a[0].Equals("Player")|| a[1].Equals("Player"))
                            list.Add("DisableInput()");
                        list.Add(string.Format("Draw({1}, {0})",d.lists.Get("name", a[0], "sword", bool.TrueString),a[0]));
                        return list;
                        }, true, "attacker", "victim")
                    .Add("SetExpression", false, "victim", "'Scared'")
                    .Add("Attack",false, "attacker", "victim")
                    .Add((a,d)=>{ return d.icons.DisableAll(a[0]); },false,"victim")
                    .Add("Die", false, "victim")
                    .Add("SetExpression", false, "victim", "'Asleep'")
                    .Add((a,d)=>{ return d.icons.Enable(a[0], "Loot", "Loot", "Loot "+a[0]); },false, "victim")
                    .Add((a,d)=>{return string.Format("Sheathe({1}, {0})",d.lists.Get("name", a[0], "sword", bool.TrueString),a[0]); },false, "attacker")
                    .Add((a,d)=>{return a[0].Equals("Player")? Reset(d): a[1].Equals("Player")? new List<string>(){ "EnableInput()" }:new List<string>(); },false, "victim", "attacker"),

                new Sequence("input Put", "container")
                    .Add("WalkTo", false, "Player", "container")
                    .Add("ClearList",true)
                    .Add((a,d)=>{return d.lists.GetAddToListActions(a[0]); },false,"Player")
                    .Add((a,d)=>{return d.lists.Get(a[0])?.SelectMany(z=> d.icons.DisableAll(z["name"])).ToList(); },false,"Player")
                    .Add((a,d)=>{return d.lists.Get(a[0])?.Select(z=> d.icons.Enable(z["name"], "PutIn", "Take", "Put "+z["description"], true)).ToList(); },false,"Player")
                    .Add((a,d)=>{return string.Format("ShowList({0}, '{0}')", a[0]); },false, "Player"),

                new Sequence("input Take", "container")
                    .Add("WalkTo", false, "Player", "Camp.Chest")
                    .Add("ClearList",true)
                    .Add((a,d)=>{return d.lists.GetAddToListActions(a[0]); },false,"Camp.Chest".GetString())
                    .Add((a,d)=>{return d.lists.Get(a[0])?.SelectMany(z=> d.icons.DisableAll(z["name"])).ToList(); },false,"Camp.Chest".GetString())
                    .Add((a,d)=>{return d.lists.Get(a[0])?.Select(z=> d.icons.Enable(z["name"], "TakeOut", "Take", "Take "+z["description"], true)).ToList(); },false,"Camp.Chest".GetString())
                    .Add((a,d)=>{return string.Format("ShowList({0}, '{1}')", a[0], "Bandit`s Chest".GetString()); },false, "Camp.Chest".GetString()),

                new Sequence("take-out", "taker", "item", "container", "place")
                    .Add((a,d)=>a[0].Equals("Player")?"HideList()":string.Format("WalkTo({0}, {1}", a[0], a[1]),false, "taker", "Camp.Chest")
                    .Add("Take", true, "taker", "item", "Camp.Chest".GetString())
                    .Add("Pocket", false, "taker", "item")
                    .Add((a,d)=>d.icons.DisableAll(a[0]),false, "item")
                    .Add((a,d)=>{ d.lists.Add(a[0], d.GetItem(a[1])); },"taker", "item")
                    .Add((a,d)=>{ d.lists.Remove(a[1],d.GetItem(a[0])); }, "item","Camp.Chest".GetString())
                    .Add((a,d)=>{ return d.icons.DisableAll(a[0]); },false,"item"),

                new Sequence("put-in", "taker", "item", "container", "place")
                    .Add("HideList")
                    .Add((a,d)=>d.icons.DisableAll(a[0]),true, "item")
                    .Add((a,d)=>{ d.lists.Add(a[0], d.GetItem(a[1])); },"Camp.Chest".GetString(), "item")
                    .Add((a,d)=>{ d.lists.Remove(a[1],d.GetItem(a[0])); }, "item","taker")
                    .Add("Put", false, "taker", "item", "Camp.Chest".GetString())
                    .Add((a,d)=>{ return d.icons.DisableAll(a[0]); },false,"item"),

                new Sequence("input Pickup","item")
                    .Add("SetNarration",false,"'You cannot take this item at this time.'".GetString())
                    .Add("ShowNarration",true),

                 new Sequence("input TakeOut","item")
                    .Add("SetNarration",false,"'You cannot take this item at this time.'".GetString())
                    .Add("ShowNarration",true),

                new Sequence("input Loot", "giver")
                    .Add("WalkTo", false, "Player", "giver")
                    .Add("ClearList",true)
                    .Add((a,d)=>{d.fluents.Set("body", a[0]); }, "giver")
                    .Add((a,d)=>{return d.lists.GetAddToListActions(a[0]); },false,"giver")
                    .Add((a,d)=>{return d.lists.Get(a[0])?.SelectMany(z=> d.icons.DisableAll(z["name"])).ToList(); },false,"giver")
                    .Add((a,d)=>{return d.lists.Get(a[0])?.Select(z=> d.icons.Enable(z["name"], "Pickup", "Take", "Take "+z["description"],true)).ToList(); },false,"giver")
                    .Add((a,d)=>{return string.Format("ShowList({0}, '{0}')", a[0]); },false, "giver"),

                //change Loot
                new Sequence("Loot", "taker", "item", "giver", "place").If((d,a)=>a[0].Equals("Player"))
                    .Add("HideList", true)
                    .Add("Take", false, "taker", "item", "giver")
                    .Add("Pocket", false, "taker", "item")
                    .Add((a,d)=>d.icons.DisableAll(a[0]),false, "item")
                    .Add((a,d)=>{ d.lists.Add(a[0], d.GetItem(a[1])); },"taker", "item")
                    .Add((a,d)=>{ d.lists.Remove(a[1],d.GetItem(a[0])); }, "item", "giver")
                    .Add((a,d)=>{ return d.icons.DisableAll(a[0]); },false,"item", "giver"),

                //change Loot
                new Sequence("Loot", "taker", "item", "giver", "place").If((d,a)=>!a[0].Equals("Player"))
                    .Add("WalkTo", false, "taker", "giver")
                    .Add("Take", true, "taker", "item", "giver")
                    .Add("Pocket", false, "taker", "item")
                    .Add((a,d)=>d.icons.DisableAll(a[0]),false, "item")
                    .Add((a,d)=>{ d.lists.Add(a[0], d.GetItem(a[1])); },"taker", "item")
                    .Add((a,d)=>{ d.lists.Remove(a[1],d.GetItem(a[0])); }, "item", "giver")
                    .Add((a,d)=>{ return d.icons.DisableAll(a[0]); },false,"item", "giver"),

                //initial cutscene
                new Sequence("input Go", "Home.Door")
                    //.If((d,args)=> d.lists.Contains("Player", GetItem("PlayerCoin".GetString())))
                    .Add("WalkTo", false, "Player", "Home.Door")
                    .Add("DisableInput", true)
                    .Add((a,d)=>d.icons.DisableAll("Home.Door", "OpenInitial"))
                    .Add((a,d)=>{return string.Format("Exit({0}, {1}, true)",a[0],a[1]); },false,"Player", "Home.Door")
                    .Add((a,d)=>{return string.Format("Enter({0}, {1}, true)",a[0],a[1]); },false,"Player", "Cottage.Door")
                    .Add((a,d)=>{d.fluents.Set("PlayerLocation", "Cottage");})
                    .Add("EnableInput"),

                new Sequence("input OpenInitial", "Home.Door")
                    .If((d,args)=> !d.lists.Contains("Player", GetItem("PlayerCoin".GetString())))
                    .Add("SetLeft",false, "Player")
                    .Add("SetRight",false, "Grandmother".GetString())
                    .Add((a,d)=>"SetDialog('Come here before you leave. I have something important to tell you! [Nevermind|Ok]')".GetString(), false)
                    .Add("ShowDialog", true),

                new Sequence("input Talk", "Grandmother")
                     .Add("WalkTo", false, "Player", "Grandmother")
                    .Add("SetLeft",false, "Player")
                    .Add("SetRight",false, "Grandmother")
                    .Add((a,d)=>"SetDialog('I am growing weaker by the hour! Please, go to the market and get some medicine. This coin should be enough... [Grandma|I will not let you down, Grandma!]')".GetString(), false)
                    .Add("ShowDialog", true)
                    .Add((a,d)=>icons.DisableAll("Grandmother")),

                new Sequence("input Selected Grandma")
                .If((d)=>d.lists.Contains("Grandmother".GetString(), GetItem("PlayerCoin".GetString())))
                     .Add("HideDialog", true)
                     .Add((a,d)=> d.icons.DisableAll("PlayerCoin".GetString()))
                     .Add("Give", false, "Grandmother", "PlayerCoin", "Player")
                     .Add("Pocket", false, "Player", "PlayerCoin")
                     .Add((a,d)=> d.lists.Remove("Grandmother".GetString(), GetItem("PlayerCoin".GetString())))
                     .Add((a,d)=>d.lists.Add("Player", GetItem("PlayerCoin".GetString())))
                     .Add((a,d)=>d.icons.DisableAll("Home.Door"))
                     .Add((a,d)=> d.icons.Enable("Home.Door", "Go", "Open", "Go to the cottage".GetString(), true))
                     .Add("SetNarration", false, "'You received a coin from your grandmother. You can press i to view your inventory.'".GetString())
                     .Add("ShowNarration"),

                new Sequence("input Key Shortcut")
                .If((d)=>d.lists.Contains("Grandmother".GetString(), GetItem("PlayerCoin".GetString())))
                     .Add((a,d)=> d.icons.DisableAll("PlayerCoin".GetString()))
                     .Add((a,d)=> d.lists.Remove("Grandmother".GetString(), GetItem("PlayerCoin".GetString())))
                     .Add((a,d)=>d.lists.Add("Player", GetItem("PlayerCoin".GetString())))
                     .Add((a,d)=>d.icons.DisableAll("Home.Door"))
                     .Add((a,d)=> d.icons.Enable("Home.Door", "Go", "Open", "Go to the cottage".GetString(), true))
                     .Add("SetNarration", false, "'You received a coin from your grandmother. You can press i to view your inventory.'".GetString()),
                    
                //new Sequence("Give", "giver", "item", "receiver", "place")
                //    .Add("HideList",true)
                //    .Add((a,d)=>{ d.lists.Add(a[0], GetItem(a[1])); }, "receiver", "item")
                //    .Add((a,d)=>{ d.lists.Remove(a[0],GetItem(a[1])); }, "giver", "item")
                //    .Add((a,d)=>{ return d.icons.DisableAll(a[0]); },false,"item")
                //    .Add((a,d)=>{ return d.GetItem(a[0])["Own"]; },false,"item")
                //    .Add((a,d)=>{return (a[0].Equals("Medicine")&&a[1].Equals("Grandmother"))?Win(d):new List<string>(); },false, "item", "receiver"),

                //new Sequence("input Give", "item")
                //    .Add("SetNarration", false,"'They did not accept your item.'")
                //    .Add("ShowNarration"),

                //new Sequence("input GiveItem", "receiver")
                //    .Add("WalkTo",false, "Player", "receiver")
                //    .Add("ClearList")
                //    .Add((a,d)=>d.fluents.Set("receiver", a[0]),"receiver")
                //    .Add((a,d)=>{return d.lists.GetAddActions(a[0]); },false,"Player")
                //    .Add((a,d)=>{return d.lists.Get(a[0]).Select(z=> icons.Enable(z["name"], "Give", "Give", "Give "+z["name"],true)).ToList(); },false,"Player")
                //    .Add("ShowList",true, "Player", "'Items to give'"),

            };
        }

        public List<string> Update(List<string> list)
        {
            var doorIcons = new Dictionary<string, string>();
            doorIcons.Add("Market.Gate", "Go to the crossroads".GetString());
            doorIcons.Add("Crossroads.Gate", "Go to the market".GetString());

            var doors = list.Where(z => z.ToLower().StartsWith("walk")).Select(z => Regex.Match(z, @"(?<=walk\(Player,\s[^,]+,\s)[^,\)]+(?=,|\))").Value).Where(z => z.Length > 1).ToList();
            var close = doorIcons.Where(x => !doors.Contains(x.Key));
            //if(list.Count>0)Console.WriteLine("@"+list.Aggregate((a, v) => a + "@" + v));
            return close.SelectMany(z => icons.DisableAll(z.Key)).Union(doorIcons.Except(close).Select(z => icons.Enable(z.Key, "Open", "Open", z.Value, true))).ToList();
        }

        private List<string> Win(Database d)
        {
            d.fluents.SetTrue("GameOver");
            return new List<string>()
            {
                "SetNarration('You saved your grandmother!')".GetString(),
                "ShowNarration()"
            };
        }

        private List<string> Reset(Database d)
        {
            d.fluents.SetTrue("GameOver");
            return new List<string>()
            {
                "SetNarration('You failed!')".GetString(),
                "ShowNarration()"
            };
        }

        private Bindings GetBindings()
        {
            //bindings[value] returns the binding whose name matches the name of the value, name being the single string that comes after input
            return new Bindings(this)
                //{i} arguments are supplied at runtime. You can provide additional arguments as optional parameters of Add
                .Add("input Attack {0}", "attack(Player, {0}, {1})", (d) => d.fluents["PlayerLocation"])

                .Add("input Buy {0}", "buy(Player, {0}, {3}, {1}, {2})", (d) => d.lists.Contains("Player", "PlayerCoin".GetString()) ? "PlayerCoin".GetString() : "BanditCoin".GetString(), (d) => d.fluents["PlayerLocation"], (d) => "Merchant".GetString())

                .Add("input Give {0}", "give(Player, {0}, {1}, {2})", (d) => d.fluents["receiver"], (d) => d.fluents["PlayerLocation"])

                .Add("input Pickup {0}", "loot(Player, {0}, {1}, {2})", (d) => d.fluents["body"], (d) => d.fluents["PlayerLocation"])

                //If you do not know the arguments, put a star after the action name and the input is mapped to the corresponding action with any arguments
                //Selected is a reserved word for which the entire string is used to find the matching binding
                .Add("input Selected Reported", "report*")

                .Add("input Steal {0}", "rob(Player, {0}, {1}, {2})", (d) => d.fluents["robbed"], (d) => d.fluents["PlayerLocation"])

                .Add("input TakeOut {0}", "take-out(Player, {0}, {1}, {2})", (d) => "Chest".GetString(), (d) => d.fluents["PlayerLocation"])
                .Add("input PutIn {0}", "put-in(Player, {0}, {1}, {2})", (d) => "Chest".GetString(), (d) => d.fluents["PlayerLocation"])

                //.Add("input Open Home.Door", "walk(Player, Home, Home.Door, Cottage.Door, Cottage)".GetString())

                .Add("input Open Crossroads.Gate", "walk(Player, Crossroads, Crossroads.Gate, Market.Gate, Market)".GetString())
                .Add("input Open Market.Gate", "walk(Player, Market, Market.Gate, Crossroads.Gate, Crossroads)".GetString())

                //arrived and exited are other exceptions in which the entire string is used to find the matching binding
                .Add("input arrived Player position Cottage.Exit", "walk(Player, Cottage, Cottage.Exit, Crossroads.EastEnd, Crossroads)".GetString())
                .Add("input arrived Player position Crossroads.WestEnd", "walk(Player, Crossroads, Crossroads.WestEnd, Camp.Exit, Camp)".GetString())
                .Add("input arrived Player position Crossroads.EastEnd", "walk(Player, Crossroads, Crossroads.EastEnd, Cottage.Exit, Cottage)".GetString())
                .Add("input arrived Player position Camp.Exit", "walk(Player, Camp, Camp.Exit, Crossroads.WestEnd, Crossroads)".GetString());
        }

        public bool CharacterIsArmed(string character)
        {
            return lists.Contains(character, "sword", bool.TrueString);
        }

        private Item GetItem(string name)
        {
            if (name == null)
                return null;
            return items.SingleOrDefault(z => z["name"].Equals(name));
        }

        private List<Item> Items()
        {
            return new List<Item>() {
                new Item("Medicine".GetString()).Add("description", "Medicine".GetString()),
                new Item("PlayerCoin".GetString()).Add("description", "Coin".GetString()),
                new Item("BanditCoin".GetString()).Add("description", "Coin".GetString()),
                new Item("BanditSword".GetString()).Add("description", "Sword".GetString()).Add("sword", bool.TrueString),
                new Item("GuardSword".GetString()).Add("description","Sword".GetString()).Add("sword", bool.TrueString),
                new Item("MerchantSword".GetString()).Add("description","Sword".GetString()).Add("sword", bool.TrueString)
            };
        }
    }
}


