/*Copyright (c) Dr. Stephen Ware and the University of New Orleans. All rights reserved.
Contributors: Alireza Shirvani, Edward T. Garcia
Licensed under the Non-Profit Open Software License version 3.0. See LICENSE file in the project root for full license information.*/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Medium
{
    public class Sequence : List<Task>
    {
        int current;
        public string name { get;set; }
        public int primary = -1;
        List<string> args;
        List<string> groundedArguments;

        Func<string[], List<Argument>> index;

        Func<Database, List<string>, bool> condition;

        public Database database { get; set; }

        public Sequence(Sequence sequence)
        {
            this.name = sequence.name;
            this.groundedArguments = new List<string>(sequence.groundedArguments);
            this.primary = sequence.primary;
            this.args = new List<string>(sequence.args);
            this.index = sequence.index;
            this.condition = sequence.condition;
            this.database = sequence.database;
            foreach (var task in sequence)
            {
                Add(task);
            }
        }

        public Sequence(string name, params string[] args)
        {
            this.name = name;
            this.groundedArguments = new List<string>();
            this.args = args.ToList();
            this.condition = (a, b) => { return true; };
            this.index = (list) =>
            {
                return list.Select(z =>
                    this.args.IndexOf(z) > -1 ?
                        new Argument(this.args.IndexOf(z)) :
                        new Argument(z)).ToList();
            };
        }

        internal Sequence If(Func<Database, bool> condition)
        {
            this.condition = (a, b) => { return condition(a); };
            return this;
        }
        internal Sequence If(Func<Database,List<string>,bool> condition)
        {
            this.condition = condition;
            return this;
        }

        internal bool Satisfied(Database database, List<string> list)
        {
            try
            {
                return condition(database, list);
            }
            catch (ArgumentOutOfRangeException)
            {
                return false;
            }
        }

        public Sequence Ground(IEnumerable<string> args)
        {
            groundedArguments= args.Select(z => z.Trim()).ToList();
            return this;
        }
        
        public void Reset()
        {
            current = 0;
        }

        public bool Matches(string input)
        {
            return name.ToLower().Equals(input.ToLower());
        }

        public Sequence DontWait()
        {
            this.Last().dontwait = true;
            return this;
        }

        public Sequence Add(string text, bool isPrimary = false, params string[] args)
        {
            if (isPrimary&&this.primary==-1)
                this.primary = Count;
            Add(new Task(text, index(args)));
            return this;
        }

        public Sequence Add(System.Action<List<string>, Database> action, params string[] args)
        {
            Add(new Task(action, index(args)));
            return this;
        }


        public Sequence Add(Func<List<string>, Database, List<string>> function,bool isPrimary=false, params string[] args)
        {
            if (isPrimary && this.primary == -1)
                this.primary = Count;
            Add(new Task(function, index(args)));
            return this;
        }

        public Sequence Add(Func<List<string>, Database, string> function,bool isPrimary=false, params string[] args)
        {
            if (isPrimary && this.primary == -1)
                this.primary = Count;
            Add(new Task(function, index(args)));
            return this;
        }

        public List<UnityAction> Next()
        {
            List<string> list = null;
            do
            {
                if (current >= this.Count)
                {
                    current = 0;
                    break;
                }
                list = this[current++].Do(groundedArguments,database);
            } while (list == null || list.Count == 0 || list[0].Equals(string.Empty));
            return list?.Select(z => new UnityAction(z.Trim(), this[current - 1].dontwait)).ToList();
        }

        public override string ToString()
        {
            string format = string.Format("{0}(", name.ToLower());
            foreach (var arg in groundedArguments)
            {
                format += arg;
                format += ", ";
            }
            if (format.Contains(','))
                format = format.Substring(0, format.Length - 2);
            format += ")";
            return format;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }
}


