/*Copyright (c) Dr. Stephen Ware and the University of New Orleans. All rights reserved.
Contributors: Alireza Shirvani, Edward T. Garcia
Licensed under the Non-Profit Open Software License version 3.0. See LICENSE file in the project root for full license information.*/
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;

namespace PlotGraph
{
    public class ActionStream:IActionStream
    {
        protected string plotgraphPath;
        protected List<long> positions = new List<long>();
        public ActionStream(string plotgraphPath, string positionPath)
        {
            this.plotgraphPath = plotgraphPath;
            if (!File.Exists(positionPath) || !ReadPositions(positionPath))
            {
                //throw new Exception("No positions file to read.");
                PopulatePositions(plotgraphPath);
                File.WriteAllLines(positionPath, positions.Select(z=>z.ToString()).ToArray());
            }
        }

        protected virtual bool ReadPositions(string positionPath)
        {
            try
            {
                using (var fs = new FileStream(positionPath, FileMode.Open))
                using (var reader = new StreamReader(fs))
                {
                    string line = "";
                    while ((line=reader.ReadLine())!=null)
                    {
                        positions.Add(long.Parse(line));
                    }
                }
                return true;
            }
            catch
            {
                throw new Exception("Something went wrong with reading the positions file.");

            }
        }

        protected virtual void PopulatePositions(string plotgraphPath)
        {
            long filePosition = -1;
            for (int i = 0;; i++)
            {
                filePosition = GetNextFilePosition(i, filePosition, plotgraphPath);
                Console.Clear();
                Console.WriteLine("Populating: " + filePosition);
                if (filePosition == -1)
                    break;
                positions.Add(filePosition);
            };
        }

        protected virtual long GetNextFilePosition(int counter, long lastFilePosition, string plotgraphPath)
        {
            if (lastFilePosition == -1)
                return 0;

            const byte NEW_LINE = 0x0A;
            const int END_OF_FILE = -1;
            using (var file = File.OpenRead(plotgraphPath))
            {
                byte[] buffer = new byte[2048];
                file.Position = lastFilePosition;
                for (int i = 0; true; i++)
                {
                    var readByte = file.ReadByte();
                    if (readByte == END_OF_FILE)
                        return -1;

                    buffer[i] = (byte)readByte;
                    if (i >= 1 && buffer[i] == NEW_LINE && buffer[i - 1] == NEW_LINE)
                        return lastFilePosition + i;
                }
            }
        }

        public virtual string[] Get(int from)
        {
            const byte NEW_LINE = 0x0A;
            const int END_OF_FILE = -1;
            List<string> strings = new List<string>();
            using (var file = File.OpenRead(plotgraphPath))
            {
                byte[] buffer = new byte[2048];
                file.Position = positions[from];
                byte previousByte = 0;
                for (int i = 0; true; i++)
                {
                    var readByte = file.ReadByte();
                    if (readByte == END_OF_FILE)
                        break;

                    buffer[i] = (byte)readByte;
                    if (previousByte == NEW_LINE && buffer[i] == NEW_LINE)
                        break;
                    if (buffer[i] == NEW_LINE)
                    {
                        if (i > 0)
                            strings.Add(Encoding.ASCII.GetString(buffer, 0, i));

                        previousByte = NEW_LINE;
                        i = -1;
                    }
                    else
                        previousByte = 0;
                }
            }
            return strings.ToArray();
        }

        public int GetNextState(int from, int action)
        {
            const byte NEW_LINE = 0x0A;
            const int END_OF_FILE = -1;
            using (var file = File.OpenRead(plotgraphPath))
            {
                byte[] buffer = new byte[2048];
                file.Position = positions[from];
                for (int i = 0; true; i++)
                {
                    var readByte = file.ReadByte();
                    if (readByte == END_OF_FILE)
                        break;

                    buffer[i] = (byte)readByte;
                    if (buffer[i] == NEW_LINE)
                    {
                        string str = Encoding.ASCII.GetString(buffer, 0, i);
                        if (str.StartsWith(action + " "))
                        {
                            int nextState = -1;
                            int.TryParse(str.Substring((action + " ").Length), out nextState);
                            return nextState;
                        }
                        i = -1;
                    }
                    if (i >= 1 && buffer[i] == NEW_LINE && buffer[i - 1] == NEW_LINE)
                        return -1;
                }
            }
            return -1;
        }
    }
}


