/*Copyright (c) Dr. Stephen Ware and the University of New Orleans. All rights reserved.
Contributors: Alireza Shirvani, Edward T. Garcia
Licensed under the Non-Profit Open Software License version 3.0. See LICENSE file in the project root for full license information.*/
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace PlotGraph
{
    public class Graph
    {
        const char delimiter = ',';
        Logger log;
        ActionNameMapper mapper;
        IActionStream stream;

        Node current;
        EdgePriorityList priorities;
        
        public Graph(string plotgraph, string map, string positions, string log, string id)
        {
            priorities = new EdgePriorityList();
            this.log = new Logger(log, id);
            this.log.Write(DateTime.Now.ToLongDateString() + " " + DateTime.Now.ToShortTimeString());
            mapper = new ActionNameMapper(map);
            stream = new OptimalActionStream(plotgraph, positions);
            current = new Node(0, GetEdgesFor(0));
        }

        public int GetPriority(string action)
        {
            if (priorities.ContainsKey(action))
            {
                return priorities[action];
            }
            return 0;
        }

        public List<string> Actions()
        {
            return current.edges.Select(z => mapper[z.label]).ToList();
        }

        public int? Navigate(string name)
        {
            var next = current.edges.FirstOrDefault(z => z.from.Equals(current.name)&&mapper[z.label].ToLower().Equals(name.ToLower()))?.to;
            if (!next.HasValue)
                return null;
            current = new Node(next.Value, GetEdgesFor(next.Value));

            var actions = Actions();
            priorities.RemoveAll(priorities.Select(z=>z.Key).Except(actions).ToList());
            actions.ForEach(z => priorities.Add(z));

            log.Add(name);

            return next.Value;
        }

        private List<PlotGraphEdge> GetEdgesFor(int state)
        {
            try
            {
                var edge = stream.Get(state);
                return edge.Select(z => new PlotGraphEdge(int.Parse(z.Split(delimiter)[1].Trim()), state, int.Parse(z.Split(delimiter)[2].Trim()))).ToList();
            }
            catch (FormatException ex)
            {
                throw new Exception("There is a non-integer value in the plotgraph");
            }
        }

        class Node
        {
            public int name { get; set; }
            public List<PlotGraphEdge> edges { get; set; }

            public Node(int name, List<PlotGraphEdge> edges)
            {
                this.name = name;
                this.edges = edges;
            }
            public bool empty
            {
                get { return edges.FirstOrDefault(z => z.from == name) != null; }
            }
        }

        public void Reset(string log, string id)
        {
            this.log = new Logger(log, id);
            this.log.Write(DateTime.Now.ToLongDateString() + " " + DateTime.Now.ToShortTimeString());
            current = new Node(0, GetEdgesFor(0));
        }

        public int Current()
        {
            return current.name;
        }
    }
}


