/*Copyright (c) Dr. Stephen Ware and the University of New Orleans. All rights reserved.
Contributors: Alireza Shirvani, Edward T. Garcia
Licensed under the Non-Profit Open Software License version 3.0. See LICENSE file in the project root for full license information.*/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.IO;
using System.Text.RegularExpressions;
using System.Text;
using System.Threading.Tasks;

namespace ScriptedExperienceManager
{
    // Unity doesn't support System.Tuple, so I'm making my own, both here and in the output file
    class Pair<T1, T2>
    {
        public T1 First { get; private set; }
        public T2 Second { get; private set; }
        internal Pair(T1 first, T2 second)
        {
            First = first;
            Second = second;
        }
        public override bool Equals(object obj)
        {
            var item = obj as Pair<T1, string>;
            if (item == null) { return false; }
            return (this.First.Equals(item.First) && this.Second.Equals(item.Second));
        }
        public override int GetHashCode()
        {
            return this.GetHashCode();
        }
    }

    static class Pair
    {
        public static Pair<T1, T2> New<T1, T2>(T1 first, T2 second)
        {
            var tuple = new Pair<T1, T2>(first, second);
            return tuple;
        }
    }

    class Program
    {
        private static List<Pair<string, string[]>> tuples = new List<Pair<string, string[]>>();
        private static List<Pair<string, string[]>> variables = new List<Pair<string, string[]>>();
        private static List<string> initial = new List<string>();
        private static Dictionary<string, List<string>> rules = new Dictionary<string, List<string>>();
        private static Dictionary<string, List<string>> groundedRules = new Dictionary<string, List<string>>();
        private static Dictionary<string, List<string>> variableRules = new Dictionary<string, List<string>>();
        private static List<string> groundedActions;
        private static char[] actionDelimiters = { ' ', '(', ')', ',' };

        /** Main **/
        static void Main(string[] args)
        {
            /** ------------------ Phase 1: Read file and set local variables ------------------ **/
            TextReader reader = File.OpenText("parser-input.cs");
            string[] fileLines = reader.ReadToEnd().Split('\n');
            parseLines(fileLines);

            /** ------------------ Phase 2: Ground tuple references in rule headers ------------------ **/
            groundTuples();

            /** ------------------ Phase 3: Prepare the output file ------------------ **/

            // Create a giant list of strings and then write them all to a file at the end
            List<string> lines = new List<string>();

            // Header stuff
            lines.Add("using System;");
            lines.Add("using System.Threading;");
            lines.Add("using System.Collections.Generic;");
            lines.Add("using System.Linq;");
            lines.Add("namespace ConsoleExperienceManager{");

            // Pair class, because Unity doesn't support System.Tuple
            lines.Add("class Pair<T1, T2> {");
            lines.Add("public T1 First { get; private set; }");
            lines.Add("public T2 Second { get; private set; }");
            lines.Add("internal Pair(T1 first, T2 second) {");
            lines.Add("First = first;");
            lines.Add("Second = second; }");
            lines.Add("public override bool Equals(object obj){");
            lines.Add("var item = obj as Pair<T1, T2>;");
            lines.Add("if (item == null){ return false; }");
            lines.Add("return (this.First.Equals(item.First) && this.Second.Equals(item.Second)); }");
            lines.Add("public override int GetHashCode() { return this.GetHashCode(); }");
            lines.Add("}");
            lines.Add("static class Pair {");
            lines.Add("public static Pair<T1, T2> New<T1, T2>(T1 first, T2 second) {");
            lines.Add("var tuple = new Pair<T1, T2>(first, second);");
            lines.Add("return tuple; }}");

            // Begin main class
            lines.Add("class BuiltInCommunication : Communication{");
            lines.Add("private static List<Pair<string, string[]>> varList = new List<Pair<string, string[]>>();");
            lines.Add("private static Thread thread;");
            lines.Add("public BuiltInCommunication(string arg) : this(){}");
            lines.Add("public BuiltInCommunication() : base(null){");
            lines.Add("thread = new Thread(NotMain);");
            lines.Add("thread.Start(); }");

            // Function to check for a user-side variable with a certain value at a certain index
            lines.Add("private static bool checkVar(string key, int index, string value){");
            lines.Add("foreach (Pair<string, string[]> var in varList){");
            lines.Add("if (var.First.Equals(key) && var.Second[index].Equals(value)) return true; }");
            lines.Add("return false; }");

            // Function to remove a user-side variable from the list
            lines.Add("private static void removeVar(string key, string[] values){");
            lines.Add("bool listsEqual;");
            lines.Add("for (int i = 0; i < varList.Count; i++){");
            lines.Add("if (varList[i].First.Equals(key)){");
            lines.Add("listsEqual = true;");
            lines.Add("for (int j = 0; j < values.Length; j++){");
            lines.Add("if (values[j] != varList[i].Second[j]) listsEqual = false;}");
            lines.Add("if (listsEqual) varList.RemoveAt(i); }}}");

            // Function to ground an action with a user-side variable
            lines.Add("private static void groundActionWithVariable(Pair<string, string[]> var, string action, string line){ ");
            lines.Add("int varIndex; string groundedAction = action;");
            lines.Add("foreach (string t in action.Split(new char[]{' ', '(', ')', ',' })) {");
            lines.Add("if (t.Contains(\"[\") && t.Substring(0, t.IndexOf('[')).Replace(\"$\", \"\").Equals(var.First)){");
            lines.Add("varIndex = Convert.ToInt32(t.Substring(t.IndexOf('[') + 1, t.IndexOf(']') - t.IndexOf('[') - 1));");
            lines.Add("groundedAction = groundedAction.Replace(t, var.Second[varIndex]);}}");
            lines.Add("if (groundedAction.Contains(\"; wait\")){");
            lines.Add("Console.WriteLine(\"start \" + groundedAction.Substring(0, action.IndexOf(';')));");
            lines.Add("string arg = action.Split(' ').Last<string>();");
            lines.Add("if (arg.Equals(\"started\")) while ((line = Console.ReadLine()) != \"started \" + groundedAction.Substring(0, action.IndexOf(';')).Replace(\"\'\", \"\").Replace(\";\", \"\")) { }");
            lines.Add("else if (arg.Equals(\"succeeded\")) while ((line = Console.ReadLine()) != \"succeeded \" + action.Substring(0, action.IndexOf(';')).Replace(\"\'\", \"\").Replace(\";\", \"\")) { }");
            lines.Add("else Thread.Sleep(Int32.Parse(arg) * 1000); }");
            lines.Add("else Console.WriteLine(\"start \" + groundedAction.Replace(\"; \", \"\")); }");

            // There is a reason for this
            lines.Add("private static void NotMain(){");
            lines.Add("Main(null); }");

            // Begin Main method
            lines.Add("static void Main(string[] args){");

            // Declare special variables
            lines.Add("bool gameOver = false;");
            lines.Add("string line;");

            // Declare user variables, encoded as arrays of strings paired with a name.
            foreach (Pair<string, string[]> var in variables)
            {
                string varLine = "varList.Add(new Pair<string, string[]>(\"" + var.First + "\", new string[]{";
                foreach (string val in var.Second)
                {
                    varLine += "\"" + val + "\",";
                }
                lines.Add(varLine.Substring(0, varLine.Length - 1) + "}));");
            }

            // Unity handshake
            lines.Add("Console.WriteLine(\"Camelot v0.1\");");

            // Start all messages in the init list
            foreach (string command in initial)
            {
                lines.Add(command);
            }

            // Begin reading lines from the console
            lines.Add("while (!(line = Console.ReadLine()).StartsWith(\"error\") && !gameOver){");
            lines.Add("if(line==\"\"){}"); // need some kind of initial if-statement

            /** ------------------ Phase 4: Ground variables in rule headers ------------------ **/

            string groundedCondition;
            string[] parts;
            foreach (KeyValuePair<string, List<string>> rule in groundedRules)
            {

                // First build the conditional
                string[] conditions = rule.Key.Split('&'); // TODO: Handle or complain about &&
                groundedActions = rule.Value;

                // If the rule contains no indexing in the header, go ahead and get the lines.
                if (!conditions[0].Contains("["))
                {
                    lines.AddRange(getRuleLines(conditions[0], conditions)); // <-- This function handles forall blocks and wait messages
                }

                else
                {
                    // Find the term in the header that contains variable indexing
                    foreach (string term in conditions[0].Split(' '))
                    {
                        if (term.StartsWith("$"))
                        {
                            parts = term.Split('[');
                            int index = Convert.ToInt32(parts[1].Substring(0, parts[1].IndexOf(']')));

                            // Ground a rule for each variable that matches its type/name
                            foreach (Pair<string, string[]> variable in variables)
                            {
                                if (variable.First.Equals(parts[0].Replace("$", "")))
                                {
                                    groundedCondition = conditions[0].Replace(term, variable.Second[index]);

                                    // Replace all references to this variable in the list of actions (except in forall blocks)
                                    groundedActions = replaceTerm(term, rule.Value, variable.Second);

                                    // Now get the lines for this rule
                                    lines.AddRange(getRuleLines(groundedCondition, conditions));
                                }
                            }
                        }
                        // Note: This will only work properly if there is only ONE indexed term in the header.
                    }
                }
            }
            // Close the file
            lines.Add("}}}}");

            // Write all lines to one of these output files:
            File.WriteAllLines("..\\..\\Camelot Unity Project\\Assets\\Scripts\\API\\BuiltInCommunication.cs", lines);      // <-- Release
            //File.WriteAllLines("BuiltInCommunication.cs", lines);                                                                   // <-- Debug
        }

        /** Utilities **/

        /** ------------------ Phase 5: Ground variables in forall blocks  ------------------ **/

        private static List<string> getRuleLines(string groundedCondition, string[] allConditions)
        {
            List<string> linesToReturn = new List<string>();

            // First get the conditional line
            if (groundedCondition.EndsWith(" "))
                groundedCondition = groundedCondition.Substring(0, groundedCondition.Length - 1);

            string conditionalLine = "else if (line == \"" + groundedCondition + "\"";

            // ... and handle extra conditions
            for (int i = 1; i < allConditions.Length; i++)
            {
                // For now, we can only check equality of a variable. Future possibilities: Add 'contains'
                if (allConditions[i].Contains('='))
                {
                    string[] assParts = allConditions[i].Replace(" ", "").Split('=');
                    string varName = assParts[0].Substring(0, assParts[0].IndexOf('[')).Replace("$", "");
                    int index = Convert.ToInt32(assParts[0].Substring(assParts[0].IndexOf('[') + 1, assParts[0].IndexOf(']') - assParts[0].IndexOf('[') - 1));
                    conditionalLine += " && checkVar(\"" + varName + "\"," + index + ", \"" + assParts[1] + "\")";
                }
                else
                    Debug.WriteLine("Oops! Can't handle conditions other than equality: " + allConditions[i]);
            }
            conditionalLine += "){";
            linesToReturn.Add(conditionalLine);

            // Now get the rest of the lines
            string line, action;
            string[] parts;
            int rightIndex;
            List<string> valuesToCheck;
            for (int i = 0; i < groundedActions.Count(); i++)
            {
                action = groundedActions.ElementAt(i);

                // If it's a variable assignment, e.g. $swordPosition[0] = sheathed;
                if (action.StartsWith("$") && action.Contains("="))
                {
                    parts = action.Replace(";", "").Split('=');
                    rightIndex = Convert.ToInt32(parts[0].Substring(parts[0].IndexOf('[') + 1, parts[0].IndexOf(']') - parts[0].IndexOf('[') - 1));
                    // Find each variable by that name and assign it the value at that index
                    linesToReturn.Add("foreach (Pair<string, string[]> var in varList){ if(var.First.Equals(\"" + parts[0].Substring(1, parts[0].IndexOf('[') - 1) + "\")){ var.Second[" + rightIndex + "] = \"" + parts[1].Replace(" ", "").Replace(";", "") + "\";}}");
                }
                // If it's a variable creation, e.g. $swordPosition[sheathed];
                else if (action.StartsWith("$"))
                {
                    parts = action.Replace(";", "").Split(new char[] { '[', ']' });
                    line = "varList.Add(new Pair<string, string[]>(\"" + parts[0].Replace("$", "") + "\", new string[]{";
                    foreach (string s in parts[1].Replace(";", "").Split(','))
                        line += "\"" + s + "\",";
                    line = line.Substring(0, line.Length - 1) + "}));";
                    linesToReturn.Add(line);
                }
                // If it's a variable deletion, e.g. -$inventory[Coin];
                else if (action.StartsWith("-"))
                {
                    parts = action.Replace(";", "").Split(new char[] { '[', ']' });
                    line = "removeVar(\"" + parts[0].Replace("-", "").Replace("$", "") + "\", new string[]{";
                    foreach (string s in parts[1].Replace(";", "").Split(','))
                        line += "\"" + s + "\",";
                    line = line.Substring(0, line.Length - 1) + "});";
                    linesToReturn.Add(line);
                }
                // If it's the start of a forall block
                else if (action.StartsWith("forall"))
                {
                    // Split the forall line into the left part and the (possibly nonexistent) right part
                    parts = action.Replace("forall", "").Replace(" ", "").Split('=');
                    int openBracket = parts[0].IndexOf('[');
                    int closeBracket = parts[0].IndexOf(']');

                    // The left part is either a variable or tuple. Get its index
                    int leftIndex = Convert.ToInt32(parts[0].Substring(openBracket + 1, closeBracket - openBracket - 1));

                    // Create a list for values to check against the left side 
                    valuesToCheck = new List<string>();

                    // If there's a right part and it's indexed
                    if (parts.Length > 1 && parts[1].Contains('['))
                    {

                        // Get the index of the right part
                        rightIndex = Convert.ToInt32(parts[1].Substring(parts[1].IndexOf('[') + 1, parts[1].IndexOf(']') - parts[1].IndexOf('[') - 1));

                        // If the right part is a user-side variable, then the left part must be a tuple
                        if (parts[1].Contains('$'))
                        {
                            foreach (Pair<string, string[]> pair in tuples)
                            {
                                // For each tuple of the type I'm looking for
                                if (pair.First.Equals(parts[0].Substring(0, openBracket)))
                                {
                                    // Get the lines to use if I'm grounding this action with this tuple, but wrap them in a check against the user-side variable
                                    linesToReturn.Add("if(checkVar(\"" + parts[1].Substring(0, parts[1].IndexOf('[')).Replace("$", "") + "\", " + rightIndex + ", \"" + pair.Second[leftIndex] + "\")){");
                                    linesToReturn.AddRange(groundNextAction(pair, i));
                                    linesToReturn.Add("}");
                                }
                            }
                        }
                        // If the right part is a tuple, we can use the static check list
                        else
                        {
                            foreach (Pair<string, string[]> pair in tuples)
                            {
                                if (pair.First.Equals(parts[1].Substring(0, parts[1].IndexOf('['))))
                                {
                                    valuesToCheck.Add(pair.Second[rightIndex]);
                                }
                            }
                        }
                    }
                    // If there's a right part and it's neither a variable nor a tuple, then it's a static value, so just add that value to the check list
                    else if (parts.Length > 1)
                    {
                        valuesToCheck.Add(parts[1]);
                    }
                    // If there is no right part, leave the list empty.

                    // If the left part is a tuple (and if we didn't already handle it in the right-side-variable block), we need to check the static tuples list for all these values
                    if (!parts[0].Contains("$") && (parts.Length == 1 || !parts[1].Contains('$')))
                    {
                        // For each tuple of this type
                        foreach (Pair<string, string[]> pair in tuples)
                        {
                            if (pair.First.Equals(parts[0].Substring(0, parts[0].IndexOf('['))))
                            {
                                // If it meets the value constraint, ground the action using this Pair
                                if (valuesToCheck.Count > 0)
                                {
                                    foreach (string value in valuesToCheck)
                                        if (pair.Second[leftIndex].Equals(value))
                                            linesToReturn.AddRange(groundNextAction(pair, i));
                                }
                                else
                                    linesToReturn.AddRange(groundNextAction(pair, i));
                            }
                        }
                    }
                    // If the left part is a variable, we need to loop through the user-side variables list
                    else if (parts[0].Contains("$"))
                    {
                        linesToReturn.Add("foreach (Pair<string, string[]> var in varList){");
                        // If there are value constraints, loop through them
                        if (valuesToCheck.Count > 0)
                        {
                            foreach (string value in valuesToCheck)
                            {
                                linesToReturn.Add("if(checkVar(\"" + parts[0].Substring(1, openBracket - 1) + "\", " + leftIndex + ", \"" + value + "\")){");
                                linesToReturn.Add("groundActionWithVariable(var, \"" + groundedActions.ElementAt(i + 1) + "\", line);");
                                linesToReturn.Add("}");
                            }
                        }
                        else
                        {
                            linesToReturn.Add("groundActionWithVariable(var, \"" + groundedActions.ElementAt(i + 1) + "\", line);");
                        }
                        linesToReturn.Add("}");
                    }
                    i++; // Skip the line we just handled
                }
                // If it's GameOver
                else if (action.Contains("GameOver()"))
                    linesToReturn.Add("gameOver = true;");
                // If it's a plain vanilla action
                else if (!action.Contains('$'))
                    linesToReturn.AddRange(getStartLines(action)); // Add lines to start the action
                else
                    Debug.WriteLine("Error: Not sure what's up with this line: " + action);
            }
            linesToReturn.Add("}");
            return linesToReturn;
        }

        // Ground the next action in the list using the given tuple or variable
        private static List<string> groundNextAction(Pair<string, string[]> pair, int i)
        {
            int tupleIndex;
            string action = groundedActions.ElementAt(i + 1); // TODO LATER: Handle a block of actions after forall, instead of just one
            string groundedAction = action;

            // Replace all terms in this action that reference that variable
            foreach (string t in action.Split(actionDelimiters))
            {
                if (t.Contains('[') && t.Substring(0, t.IndexOf('[')).Replace("$", "").Equals(pair.First))
                {
                    tupleIndex = Convert.ToInt32(t.Substring(t.IndexOf('[') + 1, t.IndexOf(']') - t.IndexOf('[') - 1));
                    groundedAction = groundedAction.Replace(t, pair.Second[tupleIndex]);
                }
            }
            return getStartLines(groundedAction);
        }

        // Get start lines and wait loops for the given action
        private static List<string> getStartLines(string action)
        {
            List<string> lines = new List<string>();
            // Handle wait command
            if (action.Contains("; wait"))
            {
                lines.Add("Console.WriteLine(\"start " + action.Substring(0, action.IndexOf(';')) + "\");");
                string arg = action.Split(' ').Last<string>();
                if (arg.Equals("started"))
                    lines.Add("while ((line = Console.ReadLine()) != \"started " + action.Substring(0, action.IndexOf(';')).Replace("\'", "").Replace(";", "") + "\"){}");
                else if (arg.Equals("succeeded"))
                    lines.Add("while ((line = Console.ReadLine()) != \"succeeded " + action.Substring(0, action.IndexOf(';')).Replace("\'", "").Replace(";", "") + "\"){}");
                else
                    lines.Add("Thread.Sleep(" + Int32.Parse(arg) * 1000 + ");");
            }
            else
                lines.Add("Console.WriteLine(\"start " + action.Replace(";", "") + "\");");
            return lines;
        }

        // Replace each reference to the given term in the given list of actions with the value it references (EXCEPT those in forall blocks)
        private static List<string> replaceTerm(string term, List<string> actions, string[] values)
        {
            string termName = term.Substring(0, term.IndexOf('['));
            List<string> groundedActions = new List<string>();
            string groundedAction;
            int bracketIndex = -1;
            int startIndex = 0;
            int tupleIndex;
            string action;
            for (int i = 0; i < actions.Count; i++)
            {
                action = actions.ElementAt(i);
                groundedAction = action;
                if (action.StartsWith("forall"))
                {
                    // If it's in a forall block, we'll ground them later.
                    groundedActions.Add(groundedAction);
                    groundedActions.Add(actions.ElementAt(i + 1));
                    i++;
                }
                else if (action.Contains("["))
                {
                    foreach (string t in action.Split(actionDelimiters))
                    {
                        if (t.Contains("["))
                        {
                            bracketIndex = t.LastIndexOf('['); // 2 nested = okay! But no more than 2
                            startIndex = t.IndexOf('[') + 1;
                            if (bracketIndex + 1 == startIndex)
                                startIndex = 0;
                        }
                        if (t.Contains("[") && t.Substring(startIndex, bracketIndex - startIndex).Equals(termName))
                        {
                            tupleIndex = Convert.ToInt32(t.Substring(bracketIndex + 1, t.IndexOf(']') - bracketIndex - 1));
                            groundedAction = groundedAction.Replace(t.Substring(startIndex, t.IndexOf(']') - startIndex + 1), values[tupleIndex]);
                        }
                    }
                }
                groundedActions.Add(groundedAction);
            }
            return groundedActions;
        }

        // Parse a Tuple or Variable declaration in the form of: "tuple[list, of items]" or "$variable[list, of, items]"
        private static Pair<string, string[]> parseTuple(string s)
        {
            string name = s.Substring(0, s.IndexOf('['));
            Regex regex = new Regex(@"\[.*\]"); // This will match: '[' + any character except newline any number of times + ']'
            Match match = regex.Match(s);
            string[] value = match.Value.Substring(1, match.Value.Length - 2).Split(',');
            for (int i = 0; i < value.Length; i++) // Remove leading spaces from each item
                if (value[i].StartsWith(" "))
                    value[i] = value[i].Substring(1, value[i].Length - 1);
            return new Pair<string, string[]>(name.Replace("$", ""), value);
        }

        // Ground tuple references in rule headers
        private static void groundTuples()
        {
            List<string> uniqueTypes, groundedActions;
            string groundedHeader;
            string[] parts;
            int bracketIndex, tupleIndex;

            // Loop through the rules list
            foreach (KeyValuePair<string, List<string>> rule in rules)
            {
                // If the header contains no indexing, then we don't need to ground it now. It may still contain variables, but we'll ground them at runtime.
                if (!rule.Key.Contains("["))
                    groundedRules.Add(rule.Key, rule.Value);
                else
                {
                    uniqueTypes = new List<string>();
                    bool foundTuple = false;

                    // For each different type of tuple we find in the header...
                    foreach (string term in rule.Key.Split(actionDelimiters))
                    {
                        if (term.Contains("[") && !term.StartsWith("$"))
                        {
                            foundTuple = true;
                            parts = term.Split('[');
                            if (!uniqueTypes.Contains(parts[0]))
                            {
                                uniqueTypes.Add(parts[0]);

                                // Find all tuples with this type and create a new grounded rule for each one.
                                foreach (Pair<string, string[]> tuple in tuples)
                                {
                                    if (tuple.First.Equals(parts[0]))
                                    {
                                        groundedHeader = rule.Key;

                                        // Replace ALL indexed tuples of this type, both in the header and in the actions list
                                        foreach (string t in rule.Key.Split(actionDelimiters))
                                        {
                                            bracketIndex = t.IndexOf('[');
                                            if (t.Contains("[") && t.Substring(0, bracketIndex).Equals(parts[0]))
                                            { // Not very sound. path[] and pathway[] would match
                                                tupleIndex = Convert.ToInt32(t.Substring(bracketIndex + 1, t.IndexOf(']') - bracketIndex - 1));
                                                groundedHeader = groundedHeader.Replace(t, tuple.Second[tupleIndex]);
                                            }
                                        }
                                        if (!groundedRules.ContainsKey(groundedHeader))
                                        {
                                            groundedActions = replaceTerm(term, rule.Value, tuple.Second);

                                            // Add the new rule to the groundedRules list and keep going
                                            groundedRules.Add(groundedHeader, groundedActions);
                                        }
                                        else
                                            Debug.WriteLine("Warning: Skipped a rule because another rule with this header has already been added: " + groundedHeader);
                                    }
                                }
                            }
                        }
                    }
                    // If there were only variables in the header, save grounding for later.
                    if (!foundTuple)
                    {
                        groundedRules.Add(rule.Key, rule.Value);
                    }
                }
            }
        }

        // Parse lines one at a time
        private static void parseLines(string[] fileLines)
        {
            string line, message, header;
            List<string> actions;

            for (int i = 0; i < fileLines.Length; i++)
            {
                line = fileLines[i];
                // Ignore comments and blank lines
                if (line.StartsWith("//") || line.Length == 0 || line.Equals("\r")) { }
                // Allow block comments sort of
                else if (line.StartsWith("/**"))
                    while (!(line = fileLines[++i]).Contains("*/")) { }
                // Parse initial commands (and format them for writing)
                else if (line.StartsWith("init"))
                {
                    while (!(line = fileLines[++i]).Contains('}'))
                    {
                        message = line.Replace("\t", "").Replace("\r", "").Replace("\n", "");
                        initial.Add("Console.WriteLine(\"start " + message + "\");");
                    }
                }
                // Parse rules (separate header from actions and strip extra syntax)
                else if (line.StartsWith("when"))
                {
                    header = line.Replace("when ", "").Replace("\r", "").Replace("\n", "").Replace(" {", "").Replace("{", "");
                    actions = new List<string>();
                    while (!(line = fileLines[++i]).StartsWith("}"))
                    {
                        message = line.Replace("\t", "").Replace("\r", ""); // TODO: Maybe replace \n's too, so we can space out things inside rules... but I mainly want this for init
                        actions.Add(message);
                    }
                    rules.Add(header, actions);
                }
                // Parse variables (separate variable name from array of values)
                else if (line.StartsWith("$"))
                    variables.Add(parseTuple(line));
                // Parse tuples (separate tuple name from array of values)
                else if (line.Contains('[') && line.Contains(']')) // must come after rules block                
                    tuples.Add(parseTuple(line));
                else
                    Debug.WriteLine("Parsing error! Does not compute: " + line);
            }
        }


    }
}


